@@ -28,9 +28,9 @@
 #include <inttypes.h>
 
 smpl_buf::smpl_buf(size_t len)
-	: buf_len(len), time_start(0), time_end(0),
-	  data_start(0), data_end(0)
+	: buf_len(len)
 {
+	reset();
 	data = new uint32_t[len];
 }
 
@@ -39,6 +39,14 @@ smpl_buf::~smpl_buf()
 	delete[] data;
 }
 
+void smpl_buf::reset()
+{
+	time_start = 0;
+	time_end = 0;
+	data_start = 0;
+	data_end = 0;
+}
+
 ssize_t smpl_buf::avail_smpls(TIMESTAMP timestamp) const
 {
 	if (timestamp < time_start)
@@ -154,7 +162,7 @@ std::string smpl_buf::str_status(TIMESTAMP timestamp) const
 	return ost.str();
 }
 
-std::string smpl_buf::str_code(ssize_t code)
+std::string smpl_buf::str_code(int code)
 {
 	switch (code) {
 	case ERROR_TIMESTAMP:
@@ -166,6 +174,8 @@ std::string smpl_buf::str_code(ssize_t code)
 	case ERROR_OVERFLOW:
 		return "Sample buffer: Overrun";
 	default:
-		return "Sample buffer: Unknown error";
+		std::stringstream ss;
+		ss << "Sample buffer: Unknown error " << code;
+		return ss.str();
 	}
 }
