@@ -44,6 +44,10 @@ public:
 	smpl_buf(size_t len);
 	~smpl_buf();
 
+	/** Reset this buffer, keeps the size
+	*/
+	void reset();
+
 	/** Query number of samples available for reading
 	    @param timestamp time of first sample
 	    @return number of available samples or error
@@ -68,7 +72,7 @@ public:
 	    @param code an error code
 	    @return a formatted error string
 	*/
-	static std::string str_code(ssize_t code);
+	static std::string str_code(int code);
 
 	enum err_code {
 		ERROR_TIMESTAMP = -1,
