@@ -1,6 +1,8 @@
 /*
  * Copyright (C) 2013 Thomas Tsou <tom@tsou.cc>
  *
+ * SPDX-License-Identifier: LGPL-2.1+
+ *
  * This library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Lesser General Public
  * License as published by the Free Software Foundation; either
@@ -10,10 +12,6 @@
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  * Lesser General Public License for more details.
- *
- * You should have received a copy of the GNU Lesser General Public
- * License along with this library; if not, write to the Free Software
- * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
  */
 
 #ifdef HAVE_CONFIG_H
@@ -53,12 +51,14 @@ extern "C" {
 #include <osmocom/ctrl/control_if.h>
 #include <osmocom/vty/stats.h>
 #include <osmocom/vty/command.h>
+#include <osmocom/vty/cpu_sched_vty.h>
 
 #include "convolve.h"
 #include "convert.h"
 #include "trx_vty.h"
 #include "debug.h"
 #include "osmo_signal.h"
+#include "trx_rate_ctr.h"
 }
 
 #define DEFAULT_CONFIG_FILE	"osmo-trx.cfg"
@@ -76,7 +76,6 @@ static struct ctrl_handle *g_ctrlh;
 
 static RadioDevice *usrp;
 static RadioInterface *radio;
-static Transceiver *transceiver;
 
 /* Create radio interface
  *     The interface consists of sample rate changes, frequency shifts,
@@ -122,7 +121,7 @@ static int transc_sig_cb(unsigned int subsys, unsigned int signal,
 		     void *handler_data, void *signal_data)
 {
 	switch (signal) {
-	case S_TRANSC_STOP_REQUIRED:
+	case S_MAIN_STOP_REQUIRED:
 		gshutdown = true;
                 break;
 	default:
@@ -141,18 +140,12 @@ int makeTransceiver(struct trx_ctx *trx, RadioInterface *radio)
 {
 	VectorFIFO *fifo;
 
-	transceiver = new Transceiver(trx->cfg.base_port, trx->cfg.bind_addr,
-			      trx->cfg.remote_addr, trx->cfg.tx_sps,
-			      trx->cfg.rx_sps, trx->cfg.num_chans, GSM::Time(3,0),
-			      radio, trx->cfg.rssi_offset);
-	if (!transceiver->init(trx->cfg.filler, trx->cfg.rtsc,
-		       trx->cfg.rach_delay, trx->cfg.egprs, trx->cfg.ext_rach)) {
+	transceiver = new Transceiver(&trx->cfg, GSM::Time(3,0), radio);
+	if (!transceiver->init()) {
 		LOG(ALERT) << "Failed to initialize transceiver";
 		return -1;
 	}
 
-        transceiver->setSignalHandler(transc_sig_cb);
-
 	for (size_t i = 0; i < trx->cfg.num_chans; i++) {
 		fifo = radio->receiveFIFO(i);
 		if (fifo && transceiver->receiveFIFO(fifo, i))
@@ -172,14 +165,25 @@ static void sig_handler(int signo)
 		   action like printing */
 		return;
 
-	fprintf(stdout, "signal %d received\n", signo);
+	fprintf(stderr, "signal %d received\n", signo);
 	switch (signo) {
 	case SIGINT:
 	case SIGTERM:
-		fprintf(stdout, "shutting down\n");
+		fprintf(stderr, "shutting down\n");
 		gshutdown = true;
 		break;
 	case SIGABRT:
+		/* in case of abort, we want to obtain a talloc report and
+		 * then run default SIGABRT handler, who will generate coredump
+		 * and abort the process. abort() should do this for us after we
+		 * return, but program wouldn't exit if an external SIGABRT is
+		 * received.
+		 */
+		talloc_report(tall_trx_ctx, stderr);
+		talloc_report_full(tall_trx_ctx, stderr);
+		signal(SIGABRT, SIG_DFL);
+		raise(SIGABRT);
+		break;
 	case SIGUSR1:
 		talloc_report(tall_trx_ctx, stderr);
 		talloc_report_full(tall_trx_ctx, stderr);
@@ -236,7 +240,7 @@ static void setup_signal_handlers()
 		exit(EXIT_FAILURE);
 	}
 
-	osmo_fd_setup(&signal_ofd, sfd, BSC_FD_READ, signalfd_callback, NULL, 0);
+	osmo_fd_setup(&signal_ofd, sfd, OSMO_FD_READ, signalfd_callback, NULL, 0);
 	if (osmo_fd_register(&signal_ofd) < 0) {
 		fprintf(stderr, "osmo_fd_register() failed.\n");
 		exit(EXIT_FAILURE);
@@ -245,10 +249,13 @@ static void setup_signal_handlers()
 
 static void print_help()
 {
-	fprintf(stdout, "Options:\n"
-		"  -h, --help      This text\n"
-		"  -C, --config    Filename The config file to use\n"
-		"  -V, --version   Print the version of OsmoTRX\n"
+	printf( "Some useful options:\n"
+		"  -h, --help			This text\n"
+		"  -C, --config			Filename The config file to use\n"
+		"  -V, --version		Print the version of OsmoTRX\n"
+		"\nVTY reference generation:\n"
+		"      --vty-ref-mode MODE	VTY reference generation mode (e.g. 'expert').\n"
+		"      --vty-ref-xml		Generate the VTY reference XML output and exit.\n"
 		);
 }
 
@@ -256,7 +263,32 @@ static void print_deprecated(char opt)
 {
 	LOG(WARNING) << "Cmd line option '" << opt << "' is deprecated and will be soon removed."
 		<< " Please use VTY cfg option instead."
-		<< " All cmd line options are already being overriden by VTY options if set.";
+		<< " All cmd line options are already being overridden by VTY options if set.";
+}
+
+static void handle_long_options(const char *prog_name, const int long_option)
+{
+	static int vty_ref_mode = VTY_REF_GEN_MODE_DEFAULT;
+
+	switch (long_option) {
+	case 1:
+		vty_ref_mode = get_string_value(vty_ref_gen_mode_names, optarg);
+		if (vty_ref_mode < 0) {
+			fprintf(stderr, "%s: Unknown VTY reference generation "
+				"mode '%s'\n", prog_name, optarg);
+			exit(2);
+		}
+		break;
+	case 2:
+		fprintf(stderr, "Generating the VTY reference in mode '%s' (%s)\n",
+			get_value_string(vty_ref_gen_mode_names, vty_ref_mode),
+			get_value_string(vty_ref_gen_mode_desc, vty_ref_mode));
+		vty_dump_xml_ref_mode(stdout, (enum vty_ref_gen_mode) vty_ref_mode);
+		exit(0);
+	default:
+		fprintf(stderr, "%s: error parsing cmdline options\n", prog_name);
+		exit(2);
+	}
 }
 
 static void handle_options(int argc, char **argv, struct trx_ctx* trx)
@@ -265,10 +297,13 @@ static void handle_options(int argc, char **argv, struct trx_ctx* trx)
 	unsigned int i;
 	std::vector<std::string> rx_paths, tx_paths;
 	bool rx_paths_set = false, tx_paths_set = false;
+	static int long_option = 0;
 	static struct option long_options[] = {
 		{"help", 0, 0, 'h'},
 		{"config", 1, 0, 'C'},
 		{"version", 0, 0, 'V'},
+		{"vty-ref-mode", 1, &long_option, 1},
+		{"vty-ref-xml", 0, &long_option, 2},
 		{NULL, 0, 0, 0}
 	};
 
@@ -279,6 +314,9 @@ static void handle_options(int argc, char **argv, struct trx_ctx* trx)
 			print_help();
 			exit(0);
 			break;
+		case 0:
+			handle_long_options(argv[0], long_option);
+			break;
 		case 'a':
 			print_deprecated(option);
 			osmo_talloc_replace_string(trx, &trx->cfg.dev_args, optarg);
@@ -333,20 +371,19 @@ static void handle_options(int argc, char **argv, struct trx_ctx* trx)
 			break;
 		case 'r':
 			print_deprecated(option);
-			trx->cfg.rtsc_set = true;
 			trx->cfg.rtsc = atoi(optarg);
 			if (!trx->cfg.egprs) /* Don't override egprs which sets different filler */
 				trx->cfg.filler = FILLER_NORM_RAND;
 			break;
 		case 'A':
 			print_deprecated(option);
-			trx->cfg.rach_delay_set = true;
 			trx->cfg.rach_delay = atoi(optarg);
 			trx->cfg.filler = FILLER_ACCESS_RAND;
 			break;
 		case 'R':
 			print_deprecated(option);
 			trx->cfg.rssi_offset = atof(optarg);
+			trx->cfg.force_rssi_offset = true;
 			break;
 		case 'S':
 			print_deprecated(option);
@@ -382,6 +419,11 @@ static void handle_options(int argc, char **argv, struct trx_ctx* trx)
 		}
 	}
 
+	if (argc > optind) {
+		LOG(ERROR) << "Unsupported positional arguments on command line";
+		goto bad_config;
+	}
+
 	/* Cmd line option specific validation & setup */
 
 	if (trx->cfg.num_chans > TRX_CHAN_MAX) {
@@ -432,7 +474,9 @@ static int set_sched_rr(unsigned int prio)
 	int rc;
 	memset(&param, 0, sizeof(param));
 	param.sched_priority = prio;
-	LOG(INFO) << "Setting SCHED_RR priority " << param.sched_priority;
+	LOG(INFO) << "Setting SCHED_RR priority " << param.sched_priority
+		  << ". This setting is DEPRECATED, please use 'policy rr " << param.sched_priority
+		  << "' under the 'sched' VTY node instead.";
 	rc = sched_setscheduler(getpid(), SCHED_RR, &param);
 	if (rc != 0) {
 		LOG(ERROR) << "Config: Setting SCHED_RR failed";
@@ -441,6 +485,38 @@ static int set_sched_rr(unsigned int prio)
 	return 0;
 }
 
+static void print_simd_info(void)
+{
+#ifdef HAVE_SSE3
+	LOGP(DMAIN, LOGL_INFO, "SSE3 support compiled in");
+#ifdef HAVE___BUILTIN_CPU_SUPPORTS
+	if (__builtin_cpu_supports("sse3"))
+		LOGPC(DMAIN, LOGL_INFO, " and supported by CPU\n");
+	else
+		LOGPC(DMAIN, LOGL_INFO, ", but not supported by CPU\n");
+#else
+	LOGPC(DMAIN, LOGL_INFO, ", but runtime SIMD detection disabled\n");
+#endif
+#endif
+
+#ifdef HAVE_SSE4_1
+	LOGP(DMAIN, LOGL_INFO, "SSE4.1 support compiled in");
+#ifdef HAVE___BUILTIN_CPU_SUPPORTS
+	if (__builtin_cpu_supports("sse4.1"))
+		LOGPC(DMAIN, LOGL_INFO, " and supported by CPU\n");
+	else
+		LOGPC(DMAIN, LOGL_INFO, ", but not supported by CPU\n");
+#else
+	LOGPC(DMAIN, LOGL_INFO, ", but runtime SIMD detection disabled\n");
+#endif
+#endif
+
+#ifndef HAVE_ATOMIC_OPS
+#pragma message ("Built without atomic operation support. Using Mutex, it may affect performance!")
+	LOG(NOTICE) << "Built without atomic operation support. Using Mutex, it may affect performance!";
+#endif
+}
+
 static void print_config(struct trx_ctx *trx)
 {
 	unsigned int i;
@@ -458,10 +534,14 @@ static void print_config(struct trx_ctx *trx)
 	ost << "   EDGE support............ " << trx->cfg.egprs << std::endl;
 	ost << "   Extended RACH support... " << trx->cfg.ext_rach << std::endl;
 	ost << "   Reference............... " << trx->cfg.clock_ref << std::endl;
-	ost << "   C0 Filler Table......... " << trx->cfg.filler << std::endl;
+	ost << "   Filler Burst Type....... " << get_value_string(filler_names, trx->cfg.filler) << std::endl;
+	ost << "   Filler Burst TSC........ " << trx->cfg.rtsc << std::endl;
+	ost << "   Filler Burst RACH Delay. " << trx->cfg.rach_delay << std::endl;
 	ost << "   Multi-Carrier........... " << trx->cfg.multi_arfcn << std::endl;
-	ost << "   Tuning offset........... " << trx->cfg.offset << std::endl;
-	ost << "   RSSI to dBm offset...... " << trx->cfg.rssi_offset << std::endl;
+	ost << "   LO freq. offset......... " << trx->cfg.offset << std::endl;
+	if (trx->cfg.freq_offset_khz != 0)
+		ost << "   Tune freq. offset....... " << trx->cfg.freq_offset_khz << std::endl;
+	ost << "   RSSI to dBm offset...... " << trx->cfg.rssi_offset << (trx->cfg.force_rssi_offset ? "" : " (relative)") << std::endl;
 	ost << "   Swap channels........... " << trx->cfg.swap_channels << std::endl;
 	ost << "   Tx Antennas.............";
 	for (i = 0; i < trx->cfg.num_chans; i++) {
@@ -546,40 +626,18 @@ int main(int argc, char *argv[])
 
 	g_trx_ctx = vty_trx_ctx_alloc(tall_trx_ctx);
 
-#ifdef HAVE_SSE3
-	printf("Info: SSE3 support compiled in");
-#ifdef HAVE___BUILTIN_CPU_SUPPORTS
-	if (__builtin_cpu_supports("sse3"))
-		printf(" and supported by CPU\n");
-	else
-		printf(", but not supported by CPU\n");
-#else
-	printf(", but runtime SIMD detection disabled\n");
-#endif
-#endif
-
-#ifdef HAVE_SSE4_1
-	printf("Info: SSE4.1 support compiled in");
-#ifdef HAVE___BUILTIN_CPU_SUPPORTS
-	if (__builtin_cpu_supports("sse4.1"))
-		printf(" and supported by CPU\n");
-	else
-		printf(", but not supported by CPU\n");
-#else
-	printf(", but runtime SIMD detection disabled\n");
-#endif
-#endif
-
 	convolve_init();
 	convert_init();
 
 	osmo_init_logging2(tall_trx_ctx, &log_info);
+	log_enable_multithread();
 	osmo_stats_init(tall_trx_ctx);
 	vty_init(&g_vty_info);
+	logging_vty_add_cmds();
 	ctrl_vty_init(tall_trx_ctx);
+	osmo_cpu_sched_vty_init(tall_trx_ctx);
 	trx_vty_init(g_trx_ctx);
 
-	logging_vty_add_cmds();
 	osmo_talloc_vty_add_cmds();
 	osmo_stats_vty_add_cmds();
 
@@ -597,7 +655,7 @@ int main(int argc, char *argv[])
 	if (rc < 0)
 		exit(1);
 
-	g_ctrlh = ctrl_interface_setup(NULL, OSMO_CTRL_PORT_TRX, NULL);
+	g_ctrlh = ctrl_interface_setup_dynip(NULL, ctrl_vty_get_bind_addr(), OSMO_CTRL_PORT_TRX, NULL);
 	if (!g_ctrlh) {
 		LOG(ERROR) << "Failed to create CTRL interface.\n";
 		exit(1);
@@ -614,6 +672,7 @@ int main(int argc, char *argv[])
 			" but expect your config to break in the future.";
 	}
 
+	print_simd_info();
 	print_config(g_trx_ctx);
 
 	if (trx_validate_config(g_trx_ctx) < 0) {
@@ -626,6 +685,9 @@ int main(int argc, char *argv[])
 			return EXIT_FAILURE;
 	}
 
+	osmo_signal_register_handler(SS_MAIN, transc_sig_cb, NULL);
+	trx_rate_ctr_init(tall_trx_ctx, g_trx_ctx);
+
 	srandom(time(NULL));
 
 	if(trx_start(g_trx_ctx) < 0)
@@ -638,5 +700,6 @@ int main(int argc, char *argv[])
 
 	osmo_fd_unregister(&signal_ofd);
 	osmo_fd_close(&signal_ofd);
+	osmo_signal_unregister_handler(SS_MAIN, transc_sig_cb, NULL);
 	return 0;
 }
