@@ -38,6 +38,7 @@ extern "C" {
 /* Universal resampling parameters */
 #define NUMCHUNKS				24
 
+/* number of narrow-band virtual ARFCNs in this wide-band multi-ARFCN device */
 #define MCHANS					4
 
 RadioInterfaceMulti::RadioInterfaceMulti(RadioDevice *radio, size_t tx_sps,
@@ -69,14 +70,24 @@ void RadioInterfaceMulti::close()
 	channelizer = NULL;
 	synthesis = NULL;
 
+
+	for (std::vector<signalVector*>::iterator it = history.begin(); it != history.end(); ++it)
+		delete *it;
+
 	mReceiveFIFO.resize(0);
 	powerScaling.resize(0);
 	history.resize(0);
 	active.resize(0);
+	rx_freq_state.resize(0);
+	tx_freq_state.resize(0);
 
 	RadioInterface::close();
 }
 
+/*! we re-map the physical channels from the filter bank to logical per-TRX channels
+ *  \param[in] pchan physical channel number within the channelizer
+ *  \param[in] chans total number of narrow-band ARFCN channels
+ *  \returns logical (TRX) channel number, or -1 in case there is none */
 static int getLogicalChan(size_t pchan, size_t chans)
 {
 	switch (chans) {
@@ -111,6 +122,9 @@ static int getLogicalChan(size_t pchan, size_t chans)
 	return -1;
 }
 
+/*! do we need to frequency shift our spectrum or not?
+ *  \param chans total number of channels
+ *  \returns 1 if we need to shift; 0 if not; -1 on error */
 static int getFreqShift(size_t chans)
 {
 	switch (chans) {
@@ -148,8 +162,11 @@ bool RadioInterfaceMulti::init(int type)
 	mReceiveFIFO.resize(mChans);
 	powerScaling.resize(mChans);
 	history.resize(mChans);
+	rx_freq_state.resize(mChans);
+	tx_freq_state.resize(mChans);
 	active.resize(MCHANS, false);
 
+	/* 4 == sps */
 	inchunk = RESAMP_INRATE * 4;
 	outchunk = RESAMP_OUTRATE * 4;
 
@@ -362,57 +379,83 @@ static bool fltcmp(double a, double b)
 	return fabs(a - b) < FREQ_DELTA_LIMIT ? true : false;
 }
 
+bool RadioInterfaceMulti::verify_arfcn_consistency(double freq, size_t chan, bool tx)
+{
+	double freq_i;
+	std::string str_dir = tx ? "Tx" : "Rx";
+	std::vector<struct freq_cfg_state> &v = tx ? tx_freq_state : rx_freq_state;
+
+	for (size_t i = 0; i < mChans; i++) {
+		if (i == chan)
+			continue;
+		if (!v[i].set)
+			continue;
+
+		freq_i = v[i].freq_hz + (double) ((int)chan - (int)i) * MCBTS_SPACING;
+		if (!fltcmp(freq, freq_i)) {
+			LOGCHAN(chan, DMAIN, ERROR)
+				<< "Setting " << str_dir << " frequency " << freq
+				<< " is incompatible: already configured channel "
+				<< i << " uses frequency " << v[i].freq_hz
+				<< " (expected " << freq_i << ")";
+			return false;
+		}
+	}
+	v[chan].set = true;
+	v[chan].freq_hz = freq;
+	return true;
+}
+
 bool RadioInterfaceMulti::tuneTx(double freq, size_t chan)
 {
-  if (chan >= mChans)
-    return false;
+	double shift;
 
-  double shift = (double) getFreqShift(mChans);
+	if (chan >= mChans)
+		return false;
 
-  if (!chan)
-    return mDevice->setTxFreq(freq + shift * MCBTS_SPACING);
+	if (!verify_arfcn_consistency(freq, chan, true))
+		return false;
 
-  double center = mDevice->getTxFreq();
-  if (!fltcmp(freq, center + (double) (chan - shift) * MCBTS_SPACING)) {
-    LOG(NOTICE) << "Channel " << chan << " RF Tx frequency offset is "
-                << freq / 1e6 << " MHz";
-  }
+	if (chan == 0) {
+		shift = (double) getFreqShift(mChans);
+		return mDevice->setTxFreq(freq + shift * MCBTS_SPACING);
+	}
 
-  return true;
+	return true;
 }
 
 bool RadioInterfaceMulti::tuneRx(double freq, size_t chan)
 {
-  if (chan >= mChans)
-    return false;
+	double shift;
 
-  double shift = (double) getFreqShift(mChans);
+	if (chan >= mChans)
+		return false;
 
-  if (!chan)
-    return mDevice->setRxFreq(freq + shift * MCBTS_SPACING);
+	if (!verify_arfcn_consistency(freq, chan, false))
+		return false;
 
-  double center = mDevice->getRxFreq();
-  if (!fltcmp(freq, center + (double) (chan - shift) * MCBTS_SPACING)) {
-    LOG(NOTICE) << "Channel " << chan << " RF Rx frequency offset is "
-                << freq / 1e6 << " MHz";
-  }
+	if (chan == 0) {
+		shift = (double) getFreqShift(mChans);
+		return mDevice->setRxFreq(freq + shift * MCBTS_SPACING);
+	}
 
-  return true;
+	return true;
 }
 
 double RadioInterfaceMulti::setRxGain(double db, size_t chan)
-{
-  if (chan == 0)
-    return mDevice->setRxGain(db);
-  else
-    return mDevice->getRxGain();
-}
-
-double RadioInterfaceMulti::setTxGain(double dB, size_t chan)
 {
 	if (chan == 0)
-		return mDevice->setTxGain(dB);
+		return mDevice->setRxGain(db);
 	else
-		return mDevice->getTxGain();
+		return mDevice->getRxGain();
+}
 
+double RadioInterfaceMulti::rssiOffset(size_t chan)
+{
+	return mDevice->rssiOffset(0);
+}
+
+int RadioInterfaceMulti::setPowerAttenuation(int atten, size_t chan)
+{
+	return RadioInterface::setPowerAttenuation(atten, 0);
 }
