@@ -34,6 +34,7 @@
 #include "Resampler.h"
 
 extern "C" {
+#include <osmocom/core/panic.h>
 #include "convolve.h"
 #include "scale.h"
 #include "mult.h"
@@ -1460,25 +1461,30 @@ static signalVector *downsampleBurst(const signalVector &burst)
  * It is computed from the training sequence of each received burst,
  * by comparing the "ideal" training sequence with the actual one.
  */
-static float computeCI(const signalVector *burst, CorrelationSequence *sync,
-                       float toa, int start, complex xcorr)
+static float computeCI(const signalVector *burst, const CorrelationSequence *sync,
+                       float toa, int start, const complex &xcorr)
 {
+  const int N = sync->sequence->size();
   float S, C;
-  int ps;
-
   /* Integer position where the sequence starts */
-  ps = start + 1 - sync->sequence->size() + (int)roundf(toa);
+  const int ps = start + 1 - N + (int)roundf(toa);
 
   /* Estimate Signal power */
   S = 0.0f;
-  for (int i=0, j=ps; i<(int)sync->sequence->size(); i++,j++)
+  for (int i=0, j=ps; i<(int)N; i++,j++)
     S += (*burst)[j].norm2();
-  S /= sync->sequence->size();
+  S /= N;
 
   /* Esimate Carrier power */
-  C = xcorr.norm2() / ((sync->sequence->size() - 1) * sync->gain.abs());
+  C = xcorr.norm2() / ((N - 1) * sync->gain.abs());
 
-  /* Interference = Signal - Carrier, so C/I = C / (S - C) */
+  /* Interference = Signal - Carrier, so C/I = C / (S - C).
+   * Calculated in dB:
+   * C/I_dB = 10 * log10(C/I)
+   * C/I_dB = 10 * (1/log2(10)) * log2(C/I)
+   * C/I_dB = 10 * 0.30103 * log2(C/I)
+   * C/I_dB = 3.0103 * log2(C/I)
+   */
   return 3.0103f * log2f(C / (S - C));
 }
 
@@ -1491,7 +1497,7 @@ static float computeCI(const signalVector *burst, CorrelationSequence *sync,
  * and we run full interpolating peak detection.
  */
 static int detectBurst(const signalVector &burst,
-                       signalVector &corr, CorrelationSequence *sync,
+                       signalVector &corr, const CorrelationSequence *sync,
                        float thresh, int sps, int start, int len,
                        struct estim_burst_params *ebp)
 {
@@ -1500,12 +1506,18 @@ static int detectBurst(const signalVector &burst,
   complex xcorr;
   int rc = 1;
 
-  if (sps == 4) {
-    dec = downsampleBurst(burst);
-    corr_in = dec;
-    sps = 1;
-  } else {
+  switch (sps) {
+  case 1:
     corr_in = &burst;
+    break;
+  case 4:
+    dec = downsampleBurst(burst);
+     /* Running at the downsampled rate at this point: */
+     corr_in = dec;
+     sps = 1;
+     break;
+  default:
+     osmo_panic("%s:%d SPS %d not supported! Only 1 or 4 supported", __FILE__, __LINE__, sps);
   }
 
   /* Correlate */
@@ -1515,9 +1527,6 @@ static int detectBurst(const signalVector &burst,
     goto del_ret;
   }
 
-  /* Running at the downsampled rate at this point */
-  sps = 1;
-
   /* Peak detection - place restrictions at correlation edges */
   ebp->amp = fastPeakDetect(corr, &ebp->toa);
 
@@ -1586,7 +1595,7 @@ static int detectGeneralBurst(const signalVector &rxBurst, float thresh, int sps
   // and only report clipping if we can't demod.
   float maxAmpl = maxAmplitude(rxBurst);
   if (maxAmpl > CLIP_THRESH) {
-    LOG(DEBUG) << "max burst amplitude: " << maxAmpl << " is above the clipping threshold: " << CLIP_THRESH << std::endl;
+    LOG(INFO) << "max burst amplitude: " << maxAmpl << " is above the clipping threshold: " << CLIP_THRESH << std::endl;
     clipping = true;
   }
 
@@ -1792,15 +1801,15 @@ static SoftVector *signalToSoftVector(signalVector *dec)
  * stages.
  */
 static signalVector *demodCommon(const signalVector &burst, int sps,
-                                 complex chan, float toa)
+                                 const struct estim_burst_params *ebp)
 {
   signalVector *delay, *dec;
 
   if ((sps != 1) && (sps != 4))
     return NULL;
 
-  delay = delayVector(&burst, NULL, -toa * (float) sps);
-  scaleVector(*delay, (complex) 1.0 / chan);
+  delay = delayVector(&burst, NULL, -ebp->toa * (float) sps);
+  scaleVector(*delay, (complex) 1.0 / ebp->amp);
 
   if (sps == 1)
     return delay;
@@ -1816,13 +1825,13 @@ static signalVector *demodCommon(const signalVector &burst, int sps,
  * 4 SPS (if activated) to minimize distortion through the fractional
  * delay filters. Symbol rotation and after always operates at 1 SPS.
  */
-static SoftVector *demodGmskBurst(const signalVector &rxBurst,
-                                  int sps, complex channel, float TOA)
+static SoftVector *demodGmskBurst(const signalVector &rxBurst, int sps,
+				  const struct estim_burst_params *ebp)
 {
   SoftVector *bits;
   signalVector *dec;
 
-  dec = demodCommon(rxBurst, sps, channel, TOA);
+  dec = demodCommon(rxBurst, sps, ebp);
   if (!dec)
     return NULL;
 
@@ -1835,29 +1844,51 @@ static SoftVector *demodGmskBurst(const signalVector &rxBurst,
   return bits;
 }
 
+static float computeEdgeCI(const signalVector *rot)
+{
+  float err_pwr = 0.0f;
+  float step = 2.0f * M_PI_F / 8.0f;
+
+  for (size_t i = 8; i < rot->size() - 8; i++) {
+    /* Compute the ideal symbol */
+    complex sym = (*rot)[i];
+    float phase = step * roundf(sym.arg() / step);
+    complex ideal = complex(cos(phase), sin(phase));
+
+    /* Compute the error vector */
+    complex err = ideal - sym;
+
+    /* Accumulate power */
+    err_pwr += err.norm2();
+  }
+
+  return 3.0103f * log2f(1.0f * (rot->size() - 16) / err_pwr);
+}
+
 /*
  * Demodulate an 8-PSK burst. Prior to symbol rotation, operate at
  * 4 SPS (if activated) to minimize distortion through the fractional
  * delay filters. Symbol rotation and after always operates at 1 SPS.
  *
  * Allow 1 SPS demodulation here, but note that other parts of the
- * transceiver restrict EDGE operatoin to 4 SPS - 8-PSK distortion
+ * transceiver restrict EDGE operation to 4 SPS - 8-PSK distortion
  * through the fractional delay filters at 1 SPS renders signal
  * nearly unrecoverable.
  */
-static SoftVector *demodEdgeBurst(const signalVector &burst,
-                                  int sps, complex chan, float toa)
+static SoftVector *demodEdgeBurst(const signalVector &burst, int sps,
+                                  struct estim_burst_params *ebp)
 {
   SoftVector *bits;
   signalVector *dec, *rot, *eq;
 
-  dec = demodCommon(burst, sps, chan, toa);
+  dec = demodCommon(burst, sps, ebp);
   if (!dec)
     return NULL;
 
   /* Equalize and derotate */
   eq = convolve(dec, GSMPulse4->c0_inv, NULL, NO_DELAY);
   rot = derotateEdgeBurst(*eq, 1);
+  ebp->ci = computeEdgeCI(rot);
 
   /* Soft slice and normalize */
   bits = softSliceEdgeBurst(*rot);
@@ -1869,13 +1900,13 @@ static SoftVector *demodEdgeBurst(const signalVector &burst,
   return bits;
 }
 
-SoftVector *demodAnyBurst(const signalVector &burst, int sps, complex amp,
-                          float toa, CorrType type)
+SoftVector *demodAnyBurst(const signalVector &burst, CorrType type,
+                          int sps, struct estim_burst_params *ebp)
 {
   if (type == EDGE)
-    return demodEdgeBurst(burst, sps, amp, toa);
+    return demodEdgeBurst(burst, sps, ebp);
   else
-    return demodGmskBurst(burst, sps, amp, toa);
+    return demodGmskBurst(burst, sps, ebp);
 }
 
 bool sigProcLibSetup()
