@@ -15,7 +15,7 @@ substr() { [ -z "${2##*$1*}" ]; }
 mychroot_nocwd() {
         # LC_ALL + LANGUAGE set to avoid lots of print errors due to locale not being set inside container
         # PATH is needed to be able to reach binaries like ldconfig without logging in to root, which adds the paths to PATH.
-        # PROOT_NO_SECCOMP is requried due to proot bug #106
+        # PROOT_NO_SECCOMP is required due to proot bug #106
         LC_ALL=C LANGUAGE=C PATH="$PATH:/usr/sbin:/sbin" PROOT_NO_SECCOMP=1 proot -r "$ROOTFS" -w / -b /proc --root-id -q qemu-arm-static "$@"
 }
 
@@ -23,15 +23,8 @@ mychroot() {
         mychroot_nocwd -w / "$@"
 }
 
-base="$PWD"
-deps="$base/deps"
-inst="$deps/install"
-export deps inst
-
 if [ -z "${INSIDE_CHROOT}" ]; then
 
-        osmo-clean-workspace.sh
-
         # Only use ARM chroot if host is not ARM and the target is ARM:
         if ! $(substr "arm" "$(uname -m)") && [ "x${INSTR}" = "x--with-neon" -o "x${INSTR}" = "x--with-neon-vfpv4" ]; then
 
@@ -69,20 +62,33 @@ if [ -z "${INSIDE_CHROOT}" ]; then
         fi
 fi
 
+set -ex
+
+if ! [ -x "$(command -v osmo-build-dep.sh)" ]; then
+	echo "Error: We need to have scripts/osmo-deps.sh from http://git.osmocom.org/osmo-ci/ in PATH !"
+	exit 2
+fi
+
+base="$PWD"
+deps="$base/deps"
+inst="$deps/install"
+export deps inst
+
+osmo-clean-workspace.sh
+
 mkdir "$deps" || true
 
 osmo-build-dep.sh libosmocore "" "--enable-sanitize --disable-doxygen --disable-pcsc"
-osmo-build-dep.sh libusrp
+PARALLEL_MAKE="" osmo-build-dep.sh libusrp
 
 export PKG_CONFIG_PATH="$inst/lib/pkgconfig:$PKG_CONFIG_PATH"
 export LD_LIBRARY_PATH="$inst/lib"
 export PATH="$inst/bin:$PATH"
 
-CONFIG="--enable-sanitize --enable-werror --with-uhd --with-usrp1 --with-lms $INSTR"
+CONFIG="--enable-sanitize --enable-werror --with-uhd --with-usrp1 --with-lms --with-ipc $INSTR"
 
 # Additional configure options and depends
 if [ "$WITH_MANUALS" = "1" ]; then
-	osmo-build-dep.sh osmo-gsm-manuals
 	CONFIG="$CONFIG --enable-manuals"
 fi
 
@@ -100,11 +106,13 @@ autoreconf --install --force
 $MAKE $PARALLEL_MAKE
 $MAKE check \
   || cat-testlogs.sh
-DISTCHECK_CONFIGURE_FLAGS="$CONFIG" $MAKE distcheck \
+DISTCHECK_CONFIGURE_FLAGS="$CONFIG" $MAKE $PARALLEL_MAKE distcheck \
   || cat-testlogs.sh
 
 if [ "$WITH_MANUALS" = "1" ] && [ "$PUBLISH" = "1" ]; then
 	make -C "$base/doc/manuals" publish
 fi
 
+$MAKE $PARALLEL_MAKE maintainer-clean
+
 osmo-clean-workspace.sh
